from functools import cached_property
from typing import List

import beanie
from beanie.odm import actions
from pymongo import AsyncMongoClient

from fmtr.tools import data_modelling_tools
from fmtr.tools.constants import Constants
from fmtr.tools.logging_tools import logger

ModifyEvents = [
    actions.Insert,
    actions.Replace,
    actions.Save,
    actions.SaveChanges,
    actions.Update
]


class Document(beanie.Document, data_modelling_tools.Base):
    """

    Document stub.

    """


class Client:

    def __init__(self, name, host=Constants.FMTR_DEV_HOST, port=27017, documents: List[beanie.Document] | None = None):
        self.name = name
        self.host = host
        self.port = port
        self.documents = documents

        self.client = AsyncMongoClient(self.uri, tz_aware=True)
        self.db = self.client[self.name]

    @cached_property
    def uri(self):
        return f'mongodb://{self.host}:{self.port}'

    async def connect(self):
        """

        Connect

        """
        with logger.span(f'Connecting to document database {self.uri=} {self.name=}'):
            return await beanie.init_beanie(database=self.db, document_models=self.documents)



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from python_on_whales import DockerClient
except ModuleNotFoundError as exception:
    DockerClient = MissingExtraMockModule('docker.client', exception)



import aiohasupervisor

from fmtr.tools import environment_tools as env
from fmtr.tools.ha_tools import constants


class Client(aiohasupervisor.SupervisorClient):
    """

    Client stub

    """

    def __init__(self, *args, api_url: str = constants.URL_SUPERVISOR_ADDON, token: str, **kwargs):
        token = token or env.get(constants.SUPERVISOR_TOKEN_KEY)
        super().__init__(*args, api_host=api_url, token=token, **kwargs)



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from fmtr.tools.dns_tools import server, client, dm, proxy
    import dns
except ModuleNotFoundError as exception:
    dns = server = client = dm = proxy = MissingExtraMockModule('dns', exception)



import inspect
from functools import cached_property
from typing import ClassVar, List, Any, Dict

from pydantic import BaseModel
from pydantic import RootModel, ConfigDict
from pydantic.fields import FieldInfo
from pydantic.json_schema import SkipJsonSchema
from pydantic_core import PydanticUndefined, PydanticUndefinedType

from fmtr.tools.datatype_tools import is_optional, none_else
from fmtr.tools.iterator_tools import get_class_lookup
from fmtr.tools.string_tools import camel_to_snake
from fmtr.tools.tools import Auto, Required, Empty


class Field(FieldInfo):
    """

    Allow DRYer field definitions, set annotation and defaults at the same time, easier field inheritance, etc.

    """
    NAME = Auto
    ANNOTATION = Empty
    DEFAULT = Auto
    FILLS = None
    DESCRIPTION = None
    TITLE = Auto
    SKIP_SCHEMA = False
    CONFIG = None

    def __init__(self, annotation=Empty, default=Empty, description=None, title=None, fills=None, skip_schema=None, **kwargs):
        """

        Infer default from type annotation, if enabled, use class/argument fills to create titles/descriptions, etc.

        """

        fills_super = getattr(super(), 'FILLS', None)
        self.fills = (fills_super or {}) | (self.FILLS or {}) | (fills or {})

        skip_schema = none_else(skip_schema, self.SKIP_SCHEMA)

        self.annotation = self.ANNOTATION if annotation is Empty else annotation
        if self.annotation is Empty:
            raise ValueError("Annotation must be specified.")

        if skip_schema:
            self.annotation = SkipJsonSchema[self.annotation]

        default = self.get_default_auto(default)
        if default is Required:
            default = PydanticUndefined

        description = self.get_desc(description)
        title = self.get_title_auto(title)
        kwargs |= (self.CONFIG or {})

        super().__init__(annotation=self.annotation, default=default, title=title, description=description, **kwargs)

    @classmethod
    def get_name_auto(cls) -> str:
        """

        Infer field name, if set to auto.

        """
        if cls.NAME is Auto:
            return camel_to_snake(cls.__name__)
        elif cls.NAME is None:
            return cls.__name__

        return cls.NAME

    @cached_property
    def fills(self) -> Dict[str, str]:
        """

        Get fills with filled title merged in

        """

        fills_super = getattr(super(), 'FILLS', None)

        return (fills_super or {}) | (self.FILLS or {}) | dict(title=self.get_title_auto())

    def get_default_auto(self, default) -> type[Any] | None | PydanticUndefinedType:
        """

        Infer default, if not specified.

        """

        if default is not Empty:
            return default

        if self.DEFAULT is not Auto:
            return self.DEFAULT

        if is_optional(self.annotation):
            return None
        else:
            return Required

    def get_title_auto(self, mask) -> str | None:
        """

        Get title from classname/mask

        """

        if not mask:
            mask = self.__class__.__name__ if self.TITLE is Auto else self.TITLE

        if mask:
            return mask.format(**self.fills)

        return None

    def get_desc(self, mask) -> str | None:
        """

        Fill description mask, if specified

        """

        mask = mask or self.DESCRIPTION

        if mask:
            return mask.format(**self.fills)

        return None


def to_df(*objs, name_value='value'):
    """

    DataFrame representation of Data Models as rows.

    """
    from fmtr.tools import tabular

    rows = []
    for obj in objs:
        if isinstance(obj, BaseModel):
            row = obj.model_dump()
        else:
            row = {name_value: obj}
        rows.append(row)

    df = tabular.DataFrame(rows)
    if 'id' in df.columns:
        df.set_index('id', inplace=True, drop=True)
    return df


class MixinArbitraryTypes:
    """

    Convenience for when non-serializable types are needed
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

class MixinFromJson:

    @classmethod
    def from_json(cls, json_str):
        """

        Error-tolerant deserialization

        """
        from fmtr.tools import json_fix
        data = json_fix.from_json(json_str, default={})

        if type(data) is dict:
            self = cls(**data)
        else:
            self = cls(data)

        return self


class CliRunMixin:
    """

    Mixin only so that it can also be used with Pydantic Settings.

    TODO Ideally, the run method would be defined on dm.Base and the settings base would just inherit like set.Base(BaseSettings, dm.Base), but this isn't yet tested/could break Fields.

    """

    def run(self):
        """

        Behaviour when run as a CLI command, i.e. via Pydantic Settings. Run any subcommands then exit.

        """

        from pydantic_settings import get_subcommand

        command = get_subcommand(self, is_required=False, cli_exit_on_error=False)

        if not command:
            return

        result = command.run()
        if inspect.isawaitable(result):
            import asyncio
            result = asyncio.run(result)

        raise SystemExit(result)


class Base(BaseModel, MixinFromJson, CliRunMixin):
    """

    Base model allowing model definition via a list of custom Field objects.

    """
    FIELDS: ClassVar[List[Field] | Dict[str, Field]] = []

    def __init_subclass__(cls, **kwargs):
        """

        Fetch aggregated fields metadata from the hierarchy and set annotations and FieldInfo objects in the class.

        """
        super().__init_subclass__(**kwargs)

        fields = {}
        for base in reversed(cls.__mro__):

            try:
                raw = base.FIELDS
            except AttributeError:
                raw = {}

            if isinstance(raw, dict):
                fields |= raw
            else:
                fields |= get_class_lookup(*raw, name_function=lambda cls_field: cls_field.get_name_auto())

        cls.FIELDS = fields

        for name, field in fields.items():
            if name in cls.__annotations__:
                continue

            if inspect.isclass(field):
                field = field()

            setattr(cls, name, field)
            cls.__annotations__[name] = field.annotation

    def to_df(self, **kwargs):
        """

        DataFrame representation of Data Model.

        """
        from fmtr.tools import tabular

        data = self.model_dump(**kwargs)
        df = tabular.pd.json_normalize(data, sep='_')
        return df

    @classmethod
    def to_df_empty(cls):
        """

        Empty DataFrame

        """
        from fmtr.tools import tabular

        row = {name: None for name in cls.model_fields.keys()}
        df = tabular.pd.DataFrame([row])
        return df

class Root(RootModel, MixinFromJson):
    """

    Root (list) model

    """

    def to_df(self, **kwargs):
        """

        DataFrame representation with items as rows.

        """

        """

        DataFrame representation of Data Model.

        """
        from fmtr.tools import tabular

        data = [item.model_dump(**kwargs) for item in self.items]
        dfs = [tabular.pd.json_normalize(datum, sep='_') for datum in data]
        df = tabular.pd.concat(dfs, axis=tabular.CONCAT_VERTICALLY)
        return df



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from fmtr.tools.interface_tools.interface_tools import Base, update, progress
    from fmtr.tools.interface_tools import controls
    from fmtr.tools.interface_tools.context import Context
except ModuleNotFoundError as exception:
    Interface = update = progress = controls = MissingExtraMockModule('interface', exception)



from fmtr.tools.import_tools import MissingExtraMockModule

from fmtr.tools.version_tools.version_tools import read, read_path, get

try:
    import semver

    semver = semver
    parse = semver.VersionInfo.parse

except ModuleNotFoundError as exception:
    semver = parse = MissingExtraMockModule('version.dev', exception)



from typing import ClassVar, Any

from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, YamlConfigSettingsSource, EnvSettingsSource, CliSettingsSource

from fmtr.tools import Constants
from fmtr.tools.data_modelling_tools import CliRunMixin
from fmtr.tools.path_tools import PackagePaths, Path


class YamlScriptConfigSettingsSource(YamlConfigSettingsSource):
    """

    Customer source for reading YAML *Script* (as opposed to plain YAML) configuration files.

    """

    def _read_file(self, file_path: Path) -> dict[str, Any]:
        """

        Use our own Path class to read YAML Script.

        """
        data = Path(file_path).read_yaml() or {}
        return data


class Base(BaseSettings, CliRunMixin):
    """

    Base class for settings configuration using Pydantic BaseSettings.
    Provides functionality for setting up and customizing sources for retrieving configuration values.
    Defines sources for configuration through environment variables, CLI arguments, YAML files.

    """

    ENV_NESTED_DELIMITER: ClassVar = Constants.ENV_NESTED_DELIMITER
    paths: ClassVar = PackagePaths()

    @classmethod
    def settings_customise_sources(
            cls,
            settings_cls: type[BaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """

        Define priority and additional sources. Note: the earlier items have higher priority.

        """

        sources = (
            init_settings,
            CliSettingsSource(settings_cls, cli_parse_args=True),
            EnvSettingsSource(settings_cls, env_prefix=cls.get_env_prefix(), env_nested_delimiter=cls.ENV_NESTED_DELIMITER),
            YamlScriptConfigSettingsSource(settings_cls, yaml_file=cls.paths.settings),
        )

        return sources

    @classmethod
    def get_env_prefix(cls):
        """

        Get environment variable prefix, which depends on whether the package is a namespace/singleton.

        """
        if cls.paths.is_namespace:
            stem = f'{cls.paths.org}_{cls.paths.name}'
        else:
            stem = f'{cls.paths.name}'

        prefix = f'{stem}{cls.ENV_NESTED_DELIMITER}'.upper()
        return prefix

    @property
    def version(self):
        """

        Read in version file.

        """
        from fmtr.tools import version
        return version.read_path(self.paths.version)



from typing import List, Tuple, Dict, Any, Self

import pymupdf as pm
import pymupdf4llm

from fmtr.tools import data_modelling_tools


class BoundingBox(data_modelling_tools.Base):
    left: float
    top: float
    right: float
    bottom: float

    @property
    def order(self):
        """

        Approximate natural reading order

        """
        return (self.top, self.left), (self.bottom, self.right)

    @property
    def rect(self) -> pm.Rect:
        """

        Position as a PyMuPDF Rect

        """
        return pm.Rect(self.left, self.top, self.right, self.bottom)

    @classmethod
    def from_dict(cls, data: Tuple[float]) -> Self:
        """

        Instantiate from PyMuPDF dictionary data

        """
        data = {key: value for key, value in zip(cls.model_fields.keys(), data)}
        return cls(**data)


class Span(data_modelling_tools.Base):
    size: float
    flags: int
    font: str
    color: int
    ascender: float
    descender: float
    text: str
    origin: Tuple[float, float]
    bbox: BoundingBox

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Self:
        data['bbox'] = BoundingBox.from_dict(data['bbox'])
        return cls(**data)


class Line(data_modelling_tools.Base):
    spans: List[Span]
    wmode: int
    dir: Tuple[float, float]
    bbox: BoundingBox

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'Line':
        """

        Instantiate from PyMuPDF dictionary data

        """
        data['spans'] = [Span.from_dict(span) for span in data['spans']]
        data['bbox'] = BoundingBox.from_dict(data['bbox'])
        return cls(**data)

    @property
    def text(self) -> str:
        """

        Simple text representation

        """
        return ' '.join([span.text for span in self.spans])


class Block(data_modelling_tools.Base):
    number: int
    type: int
    bbox: BoundingBox
    lines: List[Line]

    @property
    def text(self) -> str:
        """

        Simple text representation

        """
        return ' '.join([line.text for line in self.lines])

    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """

        Instantiate from PyMuPDF dictionary data

        """
        data['lines'] = [Line.from_dict(line) for line in data['lines']]
        data['bbox'] = BoundingBox.from_dict(data['bbox'])
        return cls(**data)

    @property
    def rect(self) -> pm.Rect:
        """

        Position as a PyMuPDF Rect

        """
        return self.bbox.rect


class Page(data_modelling_tools.Base):
    number: int
    width: float
    height: float
    blocks: List[Block]

    @property
    def text(self) -> str:
        """

        Simple text representation

        """
        return ' '.join([block.text for block in self.blocks])

    @classmethod
    def from_dict(cls, data: Dict) -> Self:
        """

        Instantiate from PyMuPDF dictionary data

        """

        data['blocks'] = [Block.from_dict(block) for block in data['blocks']]
        return cls(**data)

class Document(pm.Document):
    """

    Subclassed Document object with data-modelled elements property and markdown conversion.

    """

    @property
    def data(self) -> List[Page]:
        """

        Get representation of Document elements as Python objects.

        """

        pages = []

        for page_pm in self:
            data = page_pm.get_text("dict", flags=pm.TEXTFLAGS_TEXT | pm.TEXT_ACCURATE_BBOXES)
            data['number'] = page_pm.number
            page = Page.from_dict(data)
            pages.append(page)

        return pages

    def to_markdown(self, **kwargs) -> str:
        """

        Markdown output via `pymupdf4llm`

        """
        return pymupdf4llm.to_markdown(self, **kwargs)

    def to_text_pages(self) -> List[str]:
        """

        Simple text output per-page.

        """
        lines = []
        for page in self:
            text = page.get_text()
            lines.append(text)

        return lines

    def to_text(self) -> str:
        """

        Simple text output.

        """

        text = '\n'.join(self.to_text_pages())
        return text

    def split(self) -> List[Self]:
        """

        Split pages into individual documents.

        """

        documents = []
        for i, page in enumerate(self, start=1):
            document = self.__class__()
            document.insert_pdf(self, from_page=i, to_page=i)
            documents.append(document)

        return documents

if __name__ == '__main__':
    from fmtr.tools.path_tools import Path

    PATH_DATA = Path.data()
    # PATH_PDF=PATH_DATA/'chib.pdf'
    PATH_PDF = PATH_DATA / 'kvm.pdf'
    assert PATH_PDF.exists()

    doc = Document(PATH_PDF)
    data = doc.data
    md = doc.to_markdown()
    md



from types import UnionType, NoneType
from typing import Any, get_origin, get_args, Union, Annotated

from fmtr.tools.tools import Raise

TRUES = {str(True).lower(), str(1), 'y', 'yes', 'on'}


class TypeConversionFailed(ValueError):
    """

    Exception to raise for type conversion failure.

    """


def get_failure_message(raw, type_type):
    """

    Create generic type conversion failure message.

    """
    return f'Failed to convert "{raw}" (type: {type(raw)}) to type {type_type}'


def is_numeric(value) -> bool:
    """

    Test whether a variable is any numeric type

    """
    import numbers
    return isinstance(value, numbers.Number)


def to_bool(raw: Any, default=None) -> bool:
    """

    Convert a value to a Boolean

    """

    try:
        converted = str(raw).lower()
        converted = converted in TRUES
        return converted
    except ValueError as exception:
        if default is Raise:
            msg = get_failure_message(raw, bool)
            raise TypeConversionFailed(msg) from exception
        else:
            return default


def is_nan(value: Any) -> bool:
    """

    Import-tolerant test if nan

    """
    try:
        import pandas as pd
        return pd.isna(value)
    except ImportError:
        return False

def is_none(value: Any) -> bool:
    """

    Test if none - or nan

    """
    if is_nan(value):
        return True

    return value is None


def none_else(value: Any, default: Any) -> Any:
    """

    Ternary "if none else" function

    """
    if is_none(value):
        return default
    return value


def is_optional(annotation) -> bool:
    """

    Is type/annotation optional? todo should be in typing_tools?

    """
    if annotation is None:
        return True

    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin is Annotated:
        return is_optional(args[0])

    is_opt = (origin is UnionType or origin is Union) and NoneType in args
    return is_opt



import fmtr.tools.async_tools as aio
import fmtr.tools.database_tools as db
import fmtr.tools.dataclass_tools as dataclass
import fmtr.tools.datatype_tools as datatype
import fmtr.tools.environment_tools as env
import fmtr.tools.environment_tools as environment
import fmtr.tools.function_tools as function
import fmtr.tools.hash_tools as hash
import fmtr.tools.import_tools as import_
import fmtr.tools.inherit_tools as inherit
import fmtr.tools.iterator_tools as iterator
import fmtr.tools.json_tools as json
import fmtr.tools.logging_tools as logging
import fmtr.tools.name_tools as name
import fmtr.tools.networking_tools as net
import fmtr.tools.packaging_tools as packaging
import fmtr.tools.path_tools as path
import fmtr.tools.platform_tools as platform
import fmtr.tools.random_tools as random
import fmtr.tools.setup_tools as setup
import fmtr.tools.string_tools as string
from fmtr.tools import ai_tools as ai
from fmtr.tools import datetime_tools as dt
from fmtr.tools import dns_tools as dns
from fmtr.tools import docker_tools as docker
from fmtr.tools import ha_tools as ha
from fmtr.tools import interface_tools as interface
from fmtr.tools import version_tools as version
from fmtr.tools.constants import Constants
from fmtr.tools.import_tools import MissingExtraMockModule
from fmtr.tools.logging_tools import logger
# Submodules
from fmtr.tools.path_tools import Path, PackagePaths, AppPaths
from fmtr.tools.setup_tools import Setup, SetupPaths, Dependencies, Tools

try:
    from fmtr.tools import augmentation_tools as augmentation
except ModuleNotFoundError as exception:
    augmentation = MissingExtraMockModule('augmentation', exception)

try:
    from fmtr.tools import yaml_tools as yaml
except ModuleNotFoundError as exception:
    yaml = MissingExtraMockModule('yaml', exception)


try:
    from fmtr.tools import parallel_tools as parallel
except ModuleNotFoundError as exception:
    parallel = MissingExtraMockModule('parallel', exception)

try:
    from fmtr.tools import profiling_tools as profiling
    from fmtr.tools.profiling_tools import Timer
except ModuleNotFoundError as exception:
    profiling = Timer = MissingExtraMockModule('profiling', exception)

try:
    import fmtr.tools.process_tools as process
    from fmtr.tools.process_tools import ContextProcess
except ModuleNotFoundError as exception:
    process = ContextProcess = MissingExtraMockModule('process', exception)

try:
    from fmtr.tools import tokenization_tools as tokenization
except ModuleNotFoundError as exception:
    tokenization = MissingExtraMockModule('tokenization', exception)

try:
    from fmtr.tools import unicode_tools as unicode
except ModuleNotFoundError as exception:
    unicode = MissingExtraMockModule('unicode', exception)

try:
    from fmtr.tools import netrc_tools as netrc
except ModuleNotFoundError as exception:
    netrc = MissingExtraMockModule('netrc', exception)

try:
    from fmtr.tools import spaces_tools as spaces
except ModuleNotFoundError as exception:
    spaces = MissingExtraMockModule('spaces', exception)

try:
    from fmtr.tools import hfh_tools as hfh
except ModuleNotFoundError as exception:
    hfh = MissingExtraMockModule('hfh', exception)

try:
    from fmtr.tools import merging_tools as merging
    from fmtr.tools.merging_tools import merge
except ModuleNotFoundError as exception:
    merging = merge = MissingExtraMockModule('merging', exception)

try:
    from fmtr.tools import api_tools as api
except ModuleNotFoundError as exception:
    api = MissingExtraMockModule('api', exception)

try:
    from fmtr.tools import data_modelling_tools as dm
except ModuleNotFoundError as exception:
    dm = MissingExtraMockModule('dm', exception)

try:
    from fmtr.tools import json_fix_tools as json_fix
except ModuleNotFoundError as exception:
    json_fix = MissingExtraMockModule('json_fix', exception)

try:
    from fmtr.tools import semantic_tools as semantic
except ModuleNotFoundError as exception:
    semantic = MissingExtraMockModule('semantic', exception)

try:
    from fmtr.tools import metric_tools as metric
except ModuleNotFoundError as exception:
    metric = MissingExtraMockModule('metric', exception)

try:
    from fmtr.tools import html_tools as html
except ModuleNotFoundError as exception:
    html = MissingExtraMockModule('html', exception)

try:
    from fmtr.tools import openai_tools as openai
except ModuleNotFoundError as exception:
    openai = MissingExtraMockModule('openai', exception)

try:
    from fmtr.tools import google_api_tools as google_api
except ModuleNotFoundError as exception:
    google_api = MissingExtraMockModule('google.api', exception)

try:
    from fmtr.tools import caching_tools as caching
except ModuleNotFoundError as exception:
    caching = MissingExtraMockModule('caching', exception)

try:
    from fmtr.tools import pdf_tools as pdf
except ModuleNotFoundError as exception:
    pdf = MissingExtraMockModule('pdf', exception)

try:
    from fmtr.tools import tabular_tools as tabular
except ModuleNotFoundError as exception:
    tabular = MissingExtraMockModule('tabular', exception)

try:
    from fmtr.tools import debugging_tools as debug
except ModuleNotFoundError as exception:
    debug = MissingExtraMockModule('debug', exception)

try:
    from fmtr.tools import settings_tools as sets
except ModuleNotFoundError as exception:
    sets = MissingExtraMockModule('sets', exception)

try:
    from fmtr.tools import pattern_tools as patterns
except ModuleNotFoundError as exception:
    patterns = MissingExtraMockModule('patterns', exception)

try:
    from fmtr.tools import http_tools as http
    from fmtr.tools.http_tools import Client
except ModuleNotFoundError as exception:
    http = Client = MissingExtraMockModule('http', exception)

try:
    from fmtr.tools import webhook_tools as webhook
except ModuleNotFoundError as exception:
    webhook = MissingExtraMockModule('webhook', exception)

try:
    from fmtr.tools import mqtt_tools as mqtt
except ModuleNotFoundError as exception:
    mqtt = MissingExtraMockModule('mqtt', exception)

try:
    from fmtr.tools import av_tools as av
except ModuleNotFoundError as exception:
    av = MissingExtraMockModule('av', exception)

try:
    from fmtr.tools import youtube_tools as youtube
except ModuleNotFoundError as exception:
    youtube = MissingExtraMockModule('youtube', exception)

try:
    import pygit2 as vcs
except ModuleNotFoundError as exception:
    vcs = MissingExtraMockModule('vcs', exception)



def get_version():
    """

    Defer reading version

    """
    return version.read()



from functools import lru_cache
from typing import Any
from yaml import CDumper as Dumper
from yaml import dump

try:
    import yamlscript
except ImportError:
    raise  # Raise if even the package isn't installed, to trigger the regular missing extra exception.
except Exception as exception:
    pass  # Allow missing binary, so we can install on-demand


def install():
    """

    Installs the YAML Script runtime binary of the specified version.

    """
    import subprocess
    from fmtr.tools import logger, packaging

    version = packaging.get_version('yamlscript')
    logger.warning(f"Installing YAML Script runtime binary version {version}...")
    result = subprocess.run(f"curl https://yamlscript.org/install | VERSION={version} LIB=1 bash", shell=True, check=True)
    return result


@lru_cache
def get_module(is_auto=True):
    """

    Get the YAML Script runtime module, installing the runtime if specified

    """
    try:
        import yamlscript
    except Exception as exception:
        if not is_auto:
            raise ImportError(f'YAML Script runtime missing and {is_auto=}. Set to {True} to install.') from exception
        install()
        import yamlscript
    return yamlscript


@lru_cache
def get_interpreter():
    """

    Fetches and returns a cached instance of the YAMLScript interpreter.

    """
    module = get_module()
    interpreter = module.YAMLScript()
    return interpreter


def to_yaml(obj: Any, **kwargs) -> str:
    """

    Serialize to YAML

    """

    kwargs = dict(allow_unicode=True, Dumper=Dumper, sort_keys=False) | kwargs
    yaml_str = dump(obj, **kwargs)
    return yaml_str


def from_yaml(yaml_str: str) -> Any:
    """

    Deserialize from YAML

    """
    obj = get_interpreter().load(yaml_str)
    return obj


if __name__ == '__main__':
    data = dict(a=1, c=2, b=3)
    yaml_str = to_yaml(data)
    yaml_str



import json

from fmtr.tools.constants import Constants


def to_json(obj):
    """

    Serialise to JSON

    """
    json_str = json.dumps(obj, indent=Constants.SERIALIZATION_INDENT, ensure_ascii=False)
    return json_str


def from_json(json_str: str):
    """

    Deserialise from JSON

    """
    obj = json.loads(json_str)
    return obj



import dns
import httpx
from dataclasses import dataclass, field
from dns import rcode as dnspython_rcode, reversename as dnspython_reversename
from dns.message import Message, QueryMessage
from dns.rrset import RRset
from functools import cached_property
from typing import Self, Optional, List

from fmtr.tools.string_tools import join

TTL_CODE_DEFAULTS = {
    dnspython_rcode.NOERROR: 300,  # Successful query
    dnspython_rcode.FORMERR: 60,  # Format error
    dnspython_rcode.SERVFAIL: 10,  # Server failure
    dnspython_rcode.NXDOMAIN: 60 * 60,  # Non-existent domain
    dnspython_rcode.NOTIMP: 60,  # Not implemented
    dnspython_rcode.REFUSED: 60,  # Refused
    dnspython_rcode.YXDOMAIN: 600,  # Name exists when it should not
    dnspython_rcode.YXRRSET: 600,  # RR Set exists when it should not
    dnspython_rcode.NXRRSET: 300,  # RR Set that should exist does not
    dnspython_rcode.NOTAUTH: 60,  # Not authorized
    dnspython_rcode.NOTZONE: 60  # Name not contained in zone
}

@dataclass
class BaseDNSData:
    """

    DNS response object.

    """
    wire: bytes

    @cached_property
    def message(self) -> Message:
        return dns.message.from_wire(self.wire)

    @classmethod
    def from_message(cls, message: Message) -> Self:
        return cls(message.to_wire())

@dataclass
class Response(BaseDNSData):
    """

    DNS response object.

    """

    http: Optional[httpx.Response] = None
    blocked_by: Optional[str] = None

    @classmethod
    def from_http(cls, response: httpx.Response) -> Self:
        """

        Initialise from an HTTP response.

        """
        self = cls(response.content, http=response)
        return self

    @property
    def answer(self) -> Optional[RRset]:
        """

        Get the latest answer, if one exists.

        """
        if not self.message.answer:
            return None
        return self.message.answer[-1]

    @property
    def rcode(self) -> dnspython_rcode.Rcode:
        return self.message.rcode()

    @property
    def rcode_text(self) -> str:
        return dnspython_rcode.to_text(self.rcode)

    @property
    def ttl(self) -> int:
        """

        Get median TTL from answers, falling back to authority, then to error-code defaults.

        """
        answers = self.message.answer or self.message.authority
        if answers:
            ttls = [answer.ttl for answer in answers]
            ttl = min(ttls)
            return ttl

        ttl = TTL_CODE_DEFAULTS.get(self.rcode, dnspython_rcode.NXDOMAIN)
        return ttl



    def __str__(self):
        """

        Put answer and ID text in string representation.

        """
        answer = self.answer

        if answer:
            answer = join(answer.to_text().splitlines(), sep=', ')

        string = join([answer, self.message.flags], sep=', ')
        string = f'{self.__class__.__name__}({string})'
        return string



@dataclass
class Request(BaseDNSData):
    """

    DNS request object.

    """
    wire: bytes

    @cached_property
    def question(self) -> RRset:
        return self.message.question[0]

    @cached_property
    def is_valid(self):
        return len(self.message.question) != 0

    @cached_property
    def type(self):
        return self.question.rdtype

    @cached_property
    def type_text(self):
        return dns.rdatatype.to_text(self.type)

    @cached_property
    def name(self):
        return self.question.name

    @cached_property
    def name_text(self):
        return self.name.to_text()

    def get_response_template(self):
        message = dns.message.make_response(self.message)
        message.flags |= dns.flags.RA
        return message

    @cached_property
    def blackhole(self) -> Response:
        blackhole = self.get_response_template()
        blackhole.set_rcode(dns.rcode.NXDOMAIN)
        response = Response.from_message(blackhole)
        return response


@dataclass
class Exchange:
    """

    Entire DNS exchange for a DNS Proxy: request -> upstream response -> response

    """
    ip: str
    port: int

    request: Request
    response: Optional[Response] = None
    answers_pre: List[RRset] = field(default_factory=list)
    is_internal: bool = False
    client_name: Optional[str] = None
    is_complete: bool = False

    @property
    def addr(self):
        return self.ip, self.port

    @classmethod
    def from_wire(cls, wire: bytes, **kwargs) -> Self:
        request = Request(wire)
        response = Response.from_message(request.get_response_template())
        return cls(request=request, response=response, **kwargs)

    @cached_property
    def client(self):
        return f'{self.ip}:{self.port}'

    @property
    def question_last(self) -> RRset:
        """

        Create an RRset surrogate representing the latest/current question.
        This can be the original question - or a hybrid one if we've injected our own answers into the Exchange.
        If there's a response, use its answers, else fall back to answers_pre, else to the original question.

        """
        answers = self.answers_pre
        if self.response:
            answers = self.response.message.answer or answers

        if answers:
            rrset = answers[-1]
            rdtype = self.request.type
            ttl = self.request.question.ttl
            rdclass = self.request.question.rdclass
            name = next(iter(rrset.items.keys())).to_text()
            rrset_surrogate = dns.rrset.from_text(
                name=name,
                ttl=ttl,
                rdtype=rdtype,
                rdclass=rdclass,
            )

            return rrset_surrogate
        else:
            return self.request.question

    @property
    def query_last(self) -> QueryMessage:
        """

        Create a query (e.g. for use by upstream) based on the last question.

        """

        question_last = self.question_last
        query = dns.message.make_query(qname=question_last.name, rdclass=question_last.rdclass, rdtype=question_last.rdtype, id=self.request.message.id)
        return query

    @property
    def key(self):
        """

        Hashable key for caching

        """
        data = tuple(self.request.question.to_text().split())
        return data

    @cached_property
    def reverse(self) -> Self:
        """

        Create an Exchange for a reverse lookup of this Exchange's client IP.

        """
        name = dnspython_reversename.from_address(self.ip)
        query = dns.message.make_query(name, dns.rdatatype.PTR)
        exchange = self.__class__.from_wire(query.to_wire(), ip=self.ip, port=self.port, is_internal=True)
        return exchange



import pandas as pd


def get_peloton(data, factor: float) -> pd.Series:
    """

    Get peloton of results

    """
    top1 = data.max()
    data = data.sort_values(ascending=False)
    threshold = top1 * (1 - factor)
    peloton = data >= threshold
    return peloton



import deepdiff
import numpy as np
import pandas as pd

from fmtr.tools.iterator_tools import dedupe

Table = DataFrame = pd.DataFrame
Series = pd.Series

nan = np.nan

CONCAT_HORIZONTALLY = 1
CONCAT_VERTICALLY = 0


def normalize_nan(df, value=np.nan):
    return df.replace({pd.NA: value, None: value, np.nan: value})


class Differ:
    """

    Diff two dataframes via DeepDiff, after shape normalization, datatype simplification, etc.

    """

    def __init__(self, left: Table, right: Table):

        self.cols = dedupe(left.columns.tolist() + right.columns.tolist())
        self.rows = dedupe(left.index.values.tolist() + right.index.values.tolist())
        self.left = self.process(left)
        self.right = self.process(right)
        self.dfs = [self.left, self.right]

    def process(self, df: Table) -> Table:
        """

        Ensure same rows/columns, plus simplify datatypes via JSON round-robin.

        """

        df_rows = set(df.index.values.tolist())
        for row in self.rows:
            if row in df_rows:
                continue
            df.loc[len(df)] = None

        df_cols = set(df.columns.tolist())
        for col in self.cols:
            if col in df_cols:
                continue
            df[col] = None

        df = pd.read_json(df.to_json(date_format='iso'))
        df = normalize_nan(df, value=None)

        return df

    def get_diff(self) -> deepdiff.DeepDiff:
        """

        Cast to dicts and get diff

        """

        dicts = [df.to_dict(orient='index') for df in self.dfs]
        diff = deepdiff.DeepDiff(*dicts, ignore_order=True)
        return diff



from itertools import chain, batched

from typing import List, Dict, Any, TypeVar, Generic, Iterable

from fmtr.tools.datatype_tools import is_none


def enlist(value) -> List[Any]:
    """

    Make a non-list into a singleton list

    """
    enlisted = value if isinstance(value, list) else [value]
    return enlisted


def dict_records_to_lists(data: List[Dict[Any, Any]], missing: Any = None) -> Dict[Any, List[Any]]:
    """

    Convert a list of dictionaries to lists format

    """
    keys = set(chain.from_iterable([datum.keys() for datum in data]))
    as_lists = {key: [] for key in keys}
    for datum in data:
        for key in keys:
            as_lists[key].append(datum.get(key, missing))
    return as_lists


def get_batch_sizes(total, num_batches):
    """

    Calculate the sizes of batches for a given total number of items and number of batches.

    """
    return [total // num_batches + (1 if x < total % num_batches else 0) for x in range(num_batches)]


def chunk_data(data, size: int):
    """

    Chunk data into batches of a given size, plus any remainder

    """
    chunked = [data[offset:offset + size] for offset in range(0, len(data), size)]
    return chunked


def rebatch(batches, size: int):
    """

    Rebatch arbitrary-sized input batches into fixed-size output batches.

    """
    return batched(chain.from_iterable(batches), size)


def strip_none(*items):
    """

    Remove nones from a list of arguments

    """
    return [item for item in items if not is_none(item)]


def dedupe(items):
    """

    Deduplicate a list of items, retaining order

    """
    return list(dict.fromkeys(items))


def get_class_lookup(*classes, name_function=lambda cls: cls.__name__):
    """

    Dictionary of class names to classes

    """
    return {name_function(cls): cls for cls in classes}


IndexListT = TypeVar('IndexListT')  # Generic type for list items


class IndexList(list[IndexListT], Generic[IndexListT]):
    """

    List of objects selectable via attribute lookup, plus currently-selected item.

    """

    def __init__(self, iterable: Iterable[IndexListT] = ()):
        """

        Initialize with iterable

        """
        super().__init__(iterable)
        self.current: IndexListT | None = self[0] if self else None

    def __getattr__(self, name):
        """

        Return a lookup dict keyed on the specified field of each item in the self/list.

        """

        try:
            return self.__dict__[name]
        except KeyError:
            pass

        if hasattr(list, name):
            return getattr(self, name)

        result = {}
        for obj in self:
            try:
                value = getattr(obj, name)
            except AttributeError:
                value = obj[name]  # assume dict-like
            result[value] = obj
        return result


IterDifferT = TypeVar("IterDifferT")


class IterDiffer(Generic[IterDifferT]):
    """

    Compute added/removed differences between two iterables.

    """

    def __init__(self, before: Iterable[IterDifferT], after: Iterable[IterDifferT]):
        """

        Initialize with two iterables.

        """
        self.before: set[IterDifferT] = set(before)
        self.after: set[IterDifferT] = set(after)

    @property
    def added(self) -> set[IterDifferT]:
        """

        Items in `after` not in `before`.

        """
        return self.after - self.before

    @property
    def removed(self) -> set[IterDifferT]:
        """

        Items in `before` not in `after`.

        """
        return self.before - self.after

    @property
    def is_changed(self) -> bool:
        """

        True if any items added or removed.

        """
        return bool(self.added or self.removed)



import aiomqtt
import logging
from dataclasses import dataclass, asdict
from paho.mqtt.client import CleanStartOption, MQTT_CLEAN_START_FIRST_ONLY
from typing import Literal, Self

from fmtr.tools.logging_tools import logger, get_current_level, get_native_level_from_otel

LOGGER = logging.getLogger("mqtt")
LOGGER.handlers.clear()
LOGGER.addHandler(logger.LogfireLoggingHandler())
LOGGER.propagate = False


@dataclass
class Args:
    """

    The (serialisable subset of the) init args for Client (e.g. for init via Pydantic Settings)

    """
    hostname: str
    port: int = 1883
    username: str | None = None
    password: str | None = None
    identifier: str | None = None
    clean_session: bool | None = None
    transport: Literal["tcp", "websockets", "unix"] = "tcp"
    timeout: float | None = None
    keepalive: int = 60
    bind_address: str = ""
    bind_port: int = 0
    clean_start: CleanStartOption = MQTT_CLEAN_START_FIRST_ONLY
    max_queued_incoming_messages: int | None = None
    max_queued_outgoing_messages: int | None = None
    max_inflight_messages: int | None = None
    max_concurrent_outgoing_calls: int | None = None
    tls_insecure: bool | None = None




class Client(aiomqtt.Client):
    """

    Client stub

    """

    LOGGER = LOGGER
    SYNC_LOG_LEVEL = False
    Args = Args

    def __init__(self, *args, **kwargs):
        """

        Seems a little goofy to sync with logfire on every init, but unsure how to do it better.

        """
        if self.SYNC_LOG_LEVEL:
            self.sync_log_level()
        super().__init__(*args, **kwargs)

    def sync_log_level(self):
        """

        Sync log level with logfire, which might have changed since handler set.

        """
        level = get_current_level(logger)
        level_no = get_native_level_from_otel(level)
        self.LOGGER.setLevel(level_no)

    @classmethod
    def from_args(cls, args_obj: Args, **kwargs) -> Self:
        """

        Initialise from Args dataclass.

        """
        args = asdict(args_obj) | kwargs
        return cls(**args)

class Will(aiomqtt.Will):
    """

    Will stub

    """



from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from fmtr.tools.logging_tools import logger


class Authenticator:
    """

    Authenticate via link or saved token

    """

    PATH = None
    SCOPES = []
    SERVICE = None
    VERSION = None
    PORT = 0

    @classmethod
    def auth(cls):
        """

        Do initial authentication or refresh token if expired.

        """
        logger.info(f'Doing auth for service {cls.SERVICE} ({cls.VERSION})...')

        PATH_CREDS = cls.PATH / 'credentials.json'
        PATH_TOKEN = cls.PATH / f'{cls.SERVICE}.json'

        if PATH_TOKEN.exists():
            data_token = PATH_TOKEN.read_json()
            credentials = Credentials.from_authorized_user_info(data_token, cls.SCOPES)
            if credentials.expired:
                with logger.span(f'Credentials expired for {cls.SERVICE}. Will refresh...'):
                    logger.warning(f'{cls.SERVICE}. {PATH_CREDS.exists()=} {PATH_TOKEN.exists()=} {credentials.valid=} {credentials.expired=} {credentials.expiry=}')
                    credentials.refresh(Request())
                    PATH_TOKEN.write_text(credentials.to_json())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(PATH_CREDS, cls.SCOPES)
            flow.authorization_url(prompt='consent', access_type='offline')
            credentials = flow.run_local_server(open_browser=False, port=cls.PORT)
            PATH_TOKEN.write_text(credentials.to_json())

        return build(cls.SERVICE, cls.VERSION, credentials=credentials)



from dataclasses import dataclass
from functools import cached_property
from pytubefix import YouTube, Stream, extract, request
from pytubefix.exceptions import RegexMatchError
from typing import AsyncIterator, Iterator
from urllib.error import HTTPError

from fmtr.tools.path_tools.path_tools import Path

Stream = Stream


class Video(YouTube):
    """

    Video stub

    """


class AudioStreamDownloadError(Exception):
    """

    Error downloading audio stream

    """
    pass


@dataclass
class AudioStreamData:
    """

    Audio stream download data and progress information

    """
    message: str | None = None
    chunk: bytes | None = None
    percentage: int | None = None


class AudioStreamDownloader:
    """

    Download the highest-bitrate audio stream and write to temp directory.

    """

    def __init__(self, url_or_id: str):
        """

        Initialise with URL or video ID

        """

        try:
            self.id = extract.video_id(url_or_id)
        except RegexMatchError:
            self.id = url_or_id

        self.path = None

    @cached_property
    def url(self) -> str:
        """

        Get URL from ID

        """
        return f'https://youtube.com/watch?v={self.id}'

    async def download(self) -> AsyncIterator[AudioStreamData]:
        """

        Download the audio stream and yield chunks and progress information

        """

        yield AudioStreamData(message='Fetching video metadata...')
        video = Video(self.url)

        yield AudioStreamData('Finding audio streams...')

        audio_streams = video.streams.filter(only_audio=True).order_by('bitrate')
        if not audio_streams:
            raise AudioStreamDownloadError(f'Error downloading: no audio streams found in "{video.title}"')

        stream = audio_streams.last()
        yield AudioStreamData(f'Found highest-bitrate audio stream: {stream.audio_codec}/{stream.subtype}@{stream.abr}')

        self.path = Path.temp() / stream.default_filename
        if self.path.exists():
            self.path.unlink()

        if stream.filesize == 0:
            raise AudioStreamDownloadError(f'Error downloading: empty audio stream found in "{video.title}"')

        yield AudioStreamData('Downloading...')

        with self.path.open('wb') as out_file:
            for data in self.iter_data(stream):
                out_file.write(data.chunk)
                yield data

    def iter_data(self, stream: Stream, chunk_size: int | None = None) -> Iterator[AudioStreamData]:
        """

        Iterate over chunks of the specified size

        """
        bytes_total = bytes_remaining = stream.filesize

        if chunk_size:
            request.default_range_size = chunk_size

        try:
            stream = request.stream(stream.url)
        except HTTPError as e:
            if e.code != 404:
                raise
            stream = request.seq_stream(stream.url)

        for chunk in stream:
            bytes_remaining -= len(chunk)
            percentage = round(((bytes_total - bytes_remaining) / bytes_total) * 100)

            data = AudioStreamData(chunk=chunk, percentage=percentage)
            yield data



from typing import TypeVar, Generic

T = TypeVar("T")


class Inherit(Generic[T]):
    """

    Runtime inheritance. Acts like a wrapper around an instantiated base class of type T, and allows overriding methods in subclasses like regular inheritance.

    """

    def __init__(self, parent: T):
        """

        Set parent

        """
        object.__setattr__(self, "_parent", parent)

    def __getattr__(self, name):
        """

        Since regular attribute access checks own methods first, we don't need to do anything fancy to fall back to the parent when not implemented.

        """
        return getattr(self._parent, name)



import cachetools
from datetime import timedelta, datetime
from diskcache import Cache

from fmtr.tools.constants import Constants
from fmtr.tools.logging_tools import logger
from fmtr.tools.path_tools.path_tools import Path


class Dump(dict):
    """

    Subclass `dict` to distinguish between dumped sub-caches and regular dictionaries.

    """

class Disk(Cache):
    """

    Subclass of `diskcache` Cache that implements nested/structured caches

    """

    ROOT_KEY = '__root__'

    def __init__(self, path=None, is_root=True, **settings):
        """

        Read in existing cache structure from filesystem.

        """

        path = Path(path)
        if is_root:
            if not path.parent.exists():
                raise FileNotFoundError(f"Directory {path.parent=} does not exist")
            if path and not path.exists():
                logger.warning(f'Cache does not exist. Will be created. {str(path)=}...')

            logger.info(f'Initializing Disk Cache {str(path)=}...')

        super().__init__(directory=str(path / self.ROOT_KEY), **settings)

        self.path = path
        self.children = {}

        for path_dir in self.path.iterdir():
            if path_dir.stem == self.ROOT_KEY:
                continue
            if path_dir.is_dir():
                self.create(path_dir.name)

    def create(self, key):
        if key in self.children:
            raise KeyError(f'Sub-cache for key "{key}" already exists')
        if key in self:
            raise KeyError(f'Data for key "{key}" already exists: {repr(self[key])}')

        self.children[key] = self.__class__(self.path / key, is_root=False)

    def __getitem__(self, key):
        if key in self.children:
            return self.children[key]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if value is type(self):
            self.create(key)
        else:
            super().__setitem__(key, value)

    def setdefault(self, key, default):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return self[key]

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.path}")'

    def items(self):
        for key in self:
            yield key, self[key]

    def dump(self):
        data = Dump(self.items())
        for key, child in self.children.items():
            data[key] = child.dump()
        return data

    def iterkeys(self):
        yield from self.children.keys()
        yield from super().iterkeys()

    @property
    def data(self):
        return self.dump()


class TLRU(cachetools.TLRUCache):
    """

    Subclass to include logging and simplify global TTU

    """
    MASK_MAPPING = '{key} ' + Constants.ARROW + ' {value}'

    def __init__(self, maxsize=1_024, timer=datetime.now, getsizeof=None, ttu_static=None, desc=None):
        """

        Add overridable TTU method

        """
        super().__init__(maxsize=maxsize, ttu=self.get_ttu, timer=timer, getsizeof=getsizeof)
        self.ttu_static = ttu_static
        self.desc = desc

    @property
    def cache_desc(self):
        """

        Friendly description of cache

        """
        desc = self.desc or self.__class__.__name__
        return desc

    def get_ttu(self, _key, value, now) -> float | timedelta:
        """

        Default implementation just adds on the static TTU

        """
        return now + self.ttu_static

    def expire(self, time=None):
        """

        Log expiry

        """
        items = super().expire(time)
        if not items:
            return items

        with logger.span(f'{self.desc} cache expiry {len(items)=}...'):
            for key, value in items:
                logger.debug(self.MASK_MAPPING.format(key=key, value=value))

        return items

    def popitem(self):
        """

        Log eviction

        """
        key, value = super().popitem()
        logger.debug(f'{self.desc} cache eviction: {self.MASK_MAPPING.format(key=key, value=value)}')
        return key, value

    def dump(self):
        """

        Dump contents

        """
        data = Dump(self.items())
        return data

    @property
    def data(self):
        """

        Dump as property

        """
        return self.dump()



if __name__ == '__main__':
    sec10 = timedelta(seconds=10)
    c = TLRU(ttu_static=sec10, maxsize=2, desc='Test Data')
    c['test'] = 'val'
    c['test2'] = 'val2'
    c['test3'] = 'val3'
    c






    path_tmp_cache = Path.cwd().parent.parent / 'data' / 'cache'
    tc = Disk(path_tmp_cache)

    tc.setdefault('c', Disk).setdefault('c1', Disk)['subkey'] = 0000.1
    # tc['c']=Disk.Create
    tc['c']['test'] = False
    tc['val'] = 123
    tc.setdefault('b', Disk)
    tc.setdefault('a', Disk)
    tc['a']['value2'] = 456
    tc['a']['value4'] = dict(mykey='myvalue')

    tc['b']['value3'] = [789, True]
    tc.dump()
    {}.items()



import logging
from dataclasses import dataclass
from typing import Callable, List, Optional, Union

import uvicorn
from fastapi import FastAPI, Request

from fmtr.tools import environment_tools
from fmtr.tools.iterator_tools import enlist
from fmtr.tools.logging_tools import logger

for name in ["uvicorn.access", "uvicorn.error", "uvicorn"]:
    logger_uvicorn = logging.getLogger(name)
    logger_uvicorn.handlers.clear()
    logger_uvicorn.propagate = False

@dataclass
class Endpoint:
    """

    Endpoint-as-method config

    """
    method: Callable
    path: str
    tags: Optional[Union[str, List[str]]] = None
    method_http: Optional[Callable] = None

    def __post_init__(self):
        self.tags = enlist(self.tags)


class Base:
    """

    Simple API base class, generalising endpoint-as-method config.

    """
    TITLE = 'Base API'
    HOST = '0.0.0.0'
    PORT = 8080
    SWAGGER_PARAMS = dict(tryItOutEnabled=True)
    URL = None
    URL_DOCS = '/docs'

    def add_endpoint(self, endpoint: Endpoint):
        """

        Add endpoints from definitions using a single dataclass instance.

        """
        method_http = endpoint.method_http or self.app.post
        doc = (endpoint.method.__doc__ or '').strip() or None

        method_http(
            endpoint.path,
            tags=endpoint.tags,
            description=doc,
            summary=doc
        )(endpoint.method)

    def __init__(self):
        self.app = FastAPI(title=self.TITLE, swagger_ui_parameters=self.SWAGGER_PARAMS, docs_url=self.URL_DOCS)
        logger.instrument_fastapi(self.app)

        for endpoint in self.get_endpoints():
            self.add_endpoint(endpoint)

        if environment_tools.IS_DEV:
            self.app.exception_handler(Exception)(self.handle_exception)

    def get_endpoints(self) -> List[Endpoint]:
        """

        Define endpoints using a dataclass instance.

        """
        endpoints = [

        ]

        return endpoints

    async def handle_exception(self, request: Request, exception: Exception):
        """

        Actually raise exceptions (e.g. for debugging) instead of returning a 500.

        """
        exception
        raise

    @property
    def url(self) -> str:
        """

        Default URL unless overridden.

        """
        if self.URL:
            url = self.URL
        else:
            url = f'http://{self.HOST}:{self.PORT}'
        return url

    @property
    def message(self) -> str:
        """

        Launch message.

        """
        return f"Launching {self.TITLE} at {self.url}"

    @property
    def config(self) -> uvicorn.Config:
        """

        Uvicorn config.

        """
        return uvicorn.Config(self.app, host=self.HOST, port=self.PORT, access_log=False)

    @property
    def server(self) -> uvicorn.Server:
        """"

        Uvicorn server.

        """
        return uvicorn.Server(self.config)

    @classmethod
    async def launch_async(cls, *args, **kwargs):
        """

        Initialise and launch.

        """

        self = cls(*args, **kwargs)
        logger.info(self.message)
        await self.server.serve()

    @classmethod
    def launch(cls, *args, **kwargs):
        """

        Convenience method to launch async from a regular context.

        """
        import asyncio
        return asyncio.run(cls.launch_async(*args, **kwargs))



if __name__ == '__main__':
    Base.launch()



import asyncio
import inspect


def ensure_loop():
    """

    Ensures a loop has been started in the current thread.

    """
    try:
        return asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop


is_async = inspect.iscoroutinefunction



import socket

import uuid
from dataclasses import dataclass, field
from functools import cached_property

get_hostname = socket.gethostname
get_fqdn = socket.getfqdn


@dataclass(frozen=True)
class MAC:
    value: int = field(default_factory=uuid.getnode)

    @classmethod
    def from_string(cls, s: str):
        """

        Initialise from a string.

        """
        s = s.lower().replace(":", "").replace("-", "")
        return cls(int(s, 16))

    @cached_property
    def int(self):
        return self.value

    @cached_property
    def hex(self):
        """

        Plain 12-character hex string, no separators.

        """
        return f"{self.value:012x}"

    @cached_property
    def components(self):
        """

        Return list of six 2-digit hex strings.

        """
        h = self.hex
        return [h[i:i + 2] for i in range(0, 12, 2)]

    @cached_property
    def hex_colon(self):
        """

        Colon-separated hex string (default standard).

        """
        return self.string(":")

    @cached_property
    def is_random(self) -> bool:
        """

        Detect if this MAC is likely randomly generated by uuid.getnode().
        Checks the "multicast" bit (bit 1 of the first byte).

        """
        return bool(self.value & (1 << 40))

    def string(self, sep: str = ":") -> str:
        """

        Return MAC as a string with separator `sep`.

        """
        return sep.join(self.components)



import asyncio
from dataclasses import dataclass, field
from datetime import timedelta
from dns import rcode as dnspython_rcode
from functools import cached_property
from typing import Optional

from fmtr.tools import caching_tools as caching
from fmtr.tools.dns_tools.dm import Exchange
from fmtr.tools.logging_tools import logger


@dataclass(kw_only=True, eq=False)
class Plain(asyncio.DatagramProtocol):
    """

    Async base class for a plain DNS server using asyncio DatagramProtocol.
    """

    host: str
    port: int
    transport: Optional[asyncio.DatagramTransport] = field(default=None, init=False)

    @cached_property
    def loop(self):
        return asyncio.get_event_loop()


    @cached_property
    def cache(self):
        """

        Overridable cache.
        """
        cache = caching.TLRU(maxsize=1_024, ttu_static=timedelta(hours=1), desc='DNS Request')
        return cache

    def connection_made(self, transport: asyncio.DatagramTransport):
        self.transport = transport
        logger.info(f'Listening on {self.host}:{self.port}')

    def datagram_received(self, data: bytes, addr):
        ip, port = addr
        exchange = Exchange.from_wire(data, ip=ip, port=port)
        asyncio.create_task(self.handle(exchange))

    async def start(self):
        """

        Start the async UDP server.
        """

        logger.info(f'Starting async DNS server on {self.host}:{self.port}...')
        await self.loop.create_datagram_endpoint(
            lambda: self,
            local_addr=(self.host, self.port)
        )
        await asyncio.Future()  # Prevent exit by blocking forever

    async def resolve(self, exchange: Exchange) -> Exchange:
        """

        To be defined in subclasses.

        """
        raise NotImplementedError

    def check_cache(self, exchange: Exchange):
        if exchange.key in self.cache:
            logger.info(f'Request found in cache.')
            exchange.response = self.cache[exchange.key]
            exchange.response.message.id = exchange.request.message.id
            exchange.is_complete = True

    def get_span(self, exchange: Exchange):
        """

        Get handling span

        """
        request = exchange.request
        span = logger.span(
            f'Handling request {exchange.client_name=} {request.message.id=} {request.type_text} {request.name_text} {request.question=}...'
        )
        return span

    def log_response(self, exchange: Exchange):
        """

        Log when resolution complete

        """
        request = exchange.request
        response = exchange.response
        logger.info(
            f'Resolution complete {exchange.client_name=} {request.message.id=} {request.type_text} {request.name_text} {request.question=} {exchange.is_complete=} {response.rcode=} {response.rcode_text=} {response.answer=} {response.blocked_by=}...'
        )

    def log_dns_errors(self, exchange: Exchange):
        """

        Warn about any errors

        """
        if exchange.response.rcode != dnspython_rcode.NOERROR:
            logger.warning(f'Error {exchange.response.rcode_text=}')

    async def handle(self, exchange: Exchange):
        """

        Warn about any errors

        """
        if not exchange.request.is_valid:
            raise ValueError(f'Only one question per request is supported. Got {len(exchange.request.question)} questions.')

        if not exchange.is_internal:
            await self.handle(exchange.reverse)
            client_name = exchange.reverse.question_last.name.to_text()
            if not exchange.reverse.response.answer:
                logger.warning(f'Client name could not be resolved {client_name=}.')
            exchange.client_name = client_name

        with self.get_span(exchange):
            with logger.span(f'Checking cache...'):
                self.check_cache(exchange)

            if not exchange.is_complete:
                exchange = await self.resolve(exchange)
                self.cache[exchange.key] = exchange.response

            self.log_dns_errors(exchange)
            self.log_response(exchange)

        if exchange.is_internal:
            return

        self.transport.sendto(exchange.response.message.to_wire(), exchange.addr)



from fmtr.tools import environment_tools, Constants
from fmtr.tools.http_tools import client


def notify(title, body, url=None):
    """

    Send simple debug notification

    """
    url = url or environment_tools.get(Constants.WEBHOOK_URL_NOTIFY_KEY)
    client.post(url, json=dict(title=title, body=body))


if __name__ == '__main__':
    notify('Title', 'Body')
    notify



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from fmtr.tools.database_tools import document
except ModuleNotFoundError as exception:
    document = MissingExtraMockModule('db.document', exception)



import datetime


class Constants:
    ENCODING = 'UTF-8'
    ORG_NAME = 'fmtr'
    LIBRARY_NAME = f'{ORG_NAME}.tools'
    DATE_FILENAME_FORMAT = '%Y-%m-%d'
    TIME_FILENAME_FORMAT = '%H-%M-%S'

    DATETIME_SEMVER_BUILD_FORMAT = f'{DATE_FILENAME_FORMAT}-{TIME_FILENAME_FORMAT}'
    DATETIME_FILENAME_FORMAT = f'{DATE_FILENAME_FORMAT}@{TIME_FILENAME_FORMAT}'
    DATETIME_NOW = datetime.datetime.now(datetime.timezone.utc)
    DATETIME_NOW_STR = DATETIME_NOW.strftime(DATETIME_FILENAME_FORMAT)
    SERIALIZATION_INDENT = 4

    ENV_NESTED_DELIMITER = '__'

    ARROW = '→'
    ARROW_RIGHT = ARROW
    ARROW_SEP = f' {ARROW} '

    ARROW_LEFT = '←'

    FMTR_DEV_KEY = 'FMTR_DEV'
    FMTR_LOG_LEVEL_KEY = 'FMTR_LOG_LEVEL'
    FMTR_OBS_API_KEY_KEY = 'FMTR_OBS_API_KEY'
    FMTR_OBS_HOST = 'obs.sv.fmtr.dev'

    FMTR_REMOTE_DEBUG_HOST_KEY = 'FMTR_DEBUG_HOST'
    FMTR_REMOTE_DEBUG_HOST_DEFAULT = 'ws.lan'
    FMTR_REMOTE_DEBUG_PORT_DEFAULT = 5679
    FMTR_REMOTE_DEBUG_ENABLED_KEY = 'FMTR_REMOTE_DEBUG_ENABLED'

    FMTR_OPENAI_API_KEY_KEY = 'FMTR_OPENAI_API_KEY'

    FMTR_AI_HOST_KEY = 'FMTR_URL_HOST'
    FMTR_AI_HOST_DEFAULT = 'ai.gex.fmtr.dev'

    FMTR_DEV_HOST = 'ws.gex.fmtr.dev'

    FMTR_DEV_INTERFACE_URL = f'https://{FMTR_DEV_HOST}/'
    FMTR_DEV_INTERFACE_SUB_URL_MASK = f'https://{{sub}}.{FMTR_DEV_HOST}/'

    FILENAME_CONFIG = 'settings.yaml'
    DIR_NAME_REPO = 'repo'
    DIR_NAME_DATA = 'data'
    DIR_NAME_CACHE = 'cache'
    DIR_NAME_ARTIFACT = 'artifact'
    DIR_NAME_SOURCE = 'source'
    FILENAME_VERSION = 'version'
    DIR_NAME_HF = 'hf'

    ENTRYPOINT = 'entrypoint'
    ENTRYPOINTS_DIR = f'{ENTRYPOINT}s'
    ENTRYPOINT_FILE = f'{ENTRYPOINT}.py'
    SCRIPTS_DIR = 'scripts'

    PACKAGE_EXCLUDE_DIRS = {SCRIPTS_DIR, 'data', 'build', 'dist', '.*', '*egg-info*'}
    INIT_FILENAME = '__init__.py'
    DOCS_DIR = 'docs'
    DOCS_CONFIG_FILENAME = 'mkdocs.yml'

    DEVELOPMENT = "development"
    PRODUCTION = "production"

    INFRA = 'infra'

    PROMPT_NONE_SPECIFIED = '[None Specified]'
    WEBHOOK_URL_NOTIFY_KEY = 'WEBHOOK_URL_NOTIFY'



import logging
import os

from fmtr.tools import environment_tools
from fmtr.tools.constants import Constants

if environment_tools.IS_DEV:
    STREAM_DEFAULT = ENVIRONMENT_DEFAULT = Constants.DEVELOPMENT
else:
    STREAM_DEFAULT = None
    ENVIRONMENT_DEFAULT = Constants.PRODUCTION

LEVEL_DEFAULT = logging.DEBUG if environment_tools.IS_DEV else logging.INFO

def get_logger(name, version=None, host=Constants.FMTR_OBS_HOST, key=None, org=Constants.ORG_NAME,
               stream=STREAM_DEFAULT, environment=ENVIRONMENT_DEFAULT, level=LEVEL_DEFAULT):
    """

    Get a pre-configured logfire logger, if dependency is present, otherwise default to native logger.

    """

    stream = stream or name

    try:
        import logfire
    except ImportError:
        logger = logging.getLogger(None)
        logger.setLevel(level)
        logger.warning(f'Logging dependencies not installed. Using native logger.')

        return logger

    logger = logfire

    if key is None:
        key = environment_tools.get(Constants.FMTR_OBS_API_KEY_KEY, default=None)

    if key:
        url = f"https://{host}/api/{org}/v1/traces"
        headers = f"Authorization=Basic {key},stream-name={stream}"

        os.environ["OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"] = url
        os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = headers
        os.environ["OTEL_EXPORTER_OTLP_INSECURE"] = str(False).lower()

    if not version:
        from fmtr.tools import version_tools
        version = version_tools.read()

    lev_name_otel = get_otel_level_name(level)

    console_opts = logfire.ConsoleOptions(
        colors='always',
        min_log_level=lev_name_otel,
    )

    logfire.configure(
        service_name=name,
        service_version=version,
        environment=environment,
        send_to_logfire=False,
        console=console_opts,
        scrubbing=logfire.ScrubbingOptions(callback=null_scrubber)
    )

    return logger


def null_scrubber(match):
    """

    Effectively disable scrubbing

    """
    return match.value


def get_current_level(logger):
    """

    Get the current console log level.

    """
    level = logger.DEFAULT_LOGFIRE_INSTANCE.config.console.min_log_level
    return level


def get_logger_names():
    """

    Fetch current native logger names

    """
    return list(logging.getLogger().manager.loggerDict.keys())


OTEL_TO_NATIVE = {
    1: 1,  # trace
    3: 5,  # debug
    5: 10,  # debug (canonical)
    6: 13,  # warn
    8: 17,  # error
    9: 20,  # info
    10: 23,  # notice
    11: 25,  # success/loguru
    13: 30,  # warning
    17: 40,  # error
    21: 50,  # fatal
}


def get_otel_level_name(native_level: int) -> str:
    """

    Convert a native Python logging level number to an OTEL/logfire level name.

    """

    from logfire._internal import constants

    otel_num = constants.LOGGING_TO_OTEL_LEVEL_NUMBERS[native_level]
    name = constants.NUMBER_TO_LEVEL[otel_num]
    return name


def get_native_level_from_otel(otel_name: str) -> int:
    """

    Convert an OTEL/logfire level name to a native Python logging level number.

    """
    from logfire._internal import constants

    otel_num = constants.LEVEL_NUMBERS[otel_name]
    level = OTEL_TO_NATIVE[otel_num]
    return level


logger = get_logger(name=Constants.LIBRARY_NAME)

if __name__ == '__main__':
    logger.info('Hello World')
    logger.warning('test warning')
    logger.debug('Hello World')
    logger



from fmtr.tools.import_tools import MissingExtraMockModule

try:
    from fmtr.tools.ha_tools import core, supervisor, constants
    from fmtr.tools.ha_tools.utils import apply_addon_env

except ModuleNotFoundError as exception:
    core = supervisor = constants = apply_addon_env = MissingExtraMockModule('ha', exception)



from fmtr.tools import environment_tools as env
from fmtr.tools.constants import Constants
from fmtr.tools.inspection_tools import get_call_path


def read() -> str:
    """

    Read a generic version file from the calling package path.

    """

    path = get_call_path(offset=2).parent / Constants.FILENAME_VERSION
    return read_path(path)


def read_path(path) -> str:
    """

    Read in version from specified path

    """
    from fmtr.tools.tools import Constants
    text = path.read_text(encoding=Constants.ENCODING).strip()

    text = get(text)
    return text


def get(text) -> str:
    """

    Optionally add dev build info to raw version string.

    """

    if not env.IS_DEV:
        return text

    import datetime
    from fmtr.tools.tools import Constants
    from fmtr.tools.version_tools import parse

    timestamp = datetime.datetime.now(datetime.timezone.utc).strftime(Constants.DATETIME_SEMVER_BUILD_FORMAT)

    version = parse(text)
    version = version.bump_patch()
    version = version.replace(prerelease=Constants.DEVELOPMENT, build=timestamp)
    text = str(version)

    return text



from fmtr.tools.path_tools.path_tools import root

SUPERVISOR_TOKEN_KEY = 'SUPERVISOR_TOKEN'
URL_SUPERVISOR_ADDON = "http://supervisor"
URL_CORE_ADDON = F"{URL_SUPERVISOR_ADDON}/core/api"
PATH_ADDON_ENV = root / 'addon.env'
PATH_ADDON_OPTIONS = root / 'data' / 'options.json'
PATH_ADDON_CONFIG = root / 'config'
PATH_ADDON_MEDIA = root / 'media'



import httpx
from functools import cached_property
from httpx_retries import RetryTransport, Retry

from fmtr.tools import logging_tools

logging_tools.logger.instrument_httpx()


class Client(httpx.Client):
    """

    Instrumented client base

    """

    TIMEOUT = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, transport=self.transport, timeout=self.TIMEOUT, **kwargs)

    @cached_property
    def transport(self) -> RetryTransport:
        """

        Default Transport with retry

        """
        return RetryTransport(
            retry=self.retry
        )

    @cached_property
    def retry(self) -> Retry:
        """

        Default Retry

        """
        return Retry(
            allowed_methods=Retry.RETRYABLE_METHODS,
            backoff_factor=1.0
        )


client = Client()

if __name__ == '__main__':
    resp = client.get('https://postman-echo.com/delay/5')
    resp.raise_for_status()
    print(resp.json())
    resp



from fmtr.tools import Setup

DEPENDENCIES = {
    'dev': ['logging', 'version.dev', 'debug', 'sets', 'yaml', 'db.document'],

    'install': [],
    'test': ['pytest-cov'],
    'yaml': ['yamlscript', 'pyyaml'],
    'logging': ['logfire', 'version'],
    'parallel': ['dask[bag]', 'distributed', 'bokeh'],
    'tokenization': ['tokenizers'],
    'augmentation': ['faker', 'sre_yield'],
    'process': ['logging'],
    'profiling': ['contexttimer'],
    'docker.client': ['python-on-whales'],
    'unicode': ['Unidecode'],
    'version': [],
    'version.dev': ['semver'],
    'spaces': ['netrc'],
    'netrc': ['tinynetrc'],
    'hfh': ['huggingface_hub'],
    'merging': ['deepmerge'],
    'api': ['fastapi', 'uvicorn[standard]', 'logging', 'dm', 'logfire[fastapi]'],
    'ai': ['peft', 'transformers[sentencepiece]', 'torchvision', 'torchaudio', 'dm'],
    'dm': ['pydantic', 'pydantic-extra-types', 'pycountry'],
    'openai.api': ['openai'],
    'ai.client': ['logging', 'dm', 'openai.api', 'pydantic-ai[logfire,openai]', 'ollama'],
    'json-fix': ['json_repair'],
    'semantic': ['sentence_transformers', 'metric'],
    'metric': ['tabular'],
    'tabular': ['pandas', 'tabulate', 'openpyxl', 'odfpy', 'deepdiff'],
    'html': ['html2text'],
    'interface': ['flet[all] <0.80.0', 'flet-video', 'flet-webview'],
    'google.api': ['google-auth', 'google-auth-oauthlib', 'google-auth-httplib2', 'google-api-python-client'],
    'caching': ['diskcache', 'cachetools'],
    'pdf': ['pymupdf', 'dm', 'pymupdf4llm'],
    'debug': ['pydevd-pycharm~=251.25410.159'],
    'sets': ['pydantic-settings', 'dm', 'yaml'],
    'path.app': ['appdirs'],
    'path.type': ['filetype'],
    'dns': ['dnspython[doh]', 'http'],
    'patterns': ['regex'],
    'http': ['httpx', 'httpx_retries', 'logging', 'logfire[httpx]'],
    'setup': ['setuptools'],
    'webhook': ['http'],
    'browsers': ['playwright'],
    'db': [],
    'db.document': ['beanie[odm]'],
    'mqtt': ['aiomqtt'],
    'av': ['av'],
    'ha': ['dotenv'],
    'ha.api': ['ha', 'homeassistant_api', 'aiohasupervisor'],
    'doc': ['mkdocs', 'mkdocs-material', 'mkdocstrings[python]', 'mike', 'mkdocs-include-dir-to-nav'],
    'youtube': ['pytubefix'],
    'deploy': ['version.dev', 'logging', 'setup', 'doc', 'sets', 'build', 'twine', 'packaging', 'vcs', 'docker.client'],
    'vcs': ['pygit2'],


}

setup = Setup(
    dependencies=DEPENDENCIES,
    description='Collection of high-level tools to simplify everyday development tasks, with a focus on AI/ML',
)

setup



import os
from dotenv import load_dotenv

from fmtr.tools.ha_tools import constants
from fmtr.tools.logging_tools import logger
from fmtr.tools.string_tools import ELLIPSIS


def apply_addon_env():
    """

    If we're inside an addon container, we need to source its environment file and convert its options.json to environment variables.

    """
    path = constants.PATH_ADDON_ENV
    if path.exists():
        logger.warning(f'Loading addon environment from "{path}"...')
        load_dotenv(path)

    for key, value in convert_options_data().items():
        os.environ[key] = value


def convert_options_data() -> dict[str, str]:
    """

    Convert Home Assistant addon options.json to an environment-ready dict.

    """
    path = constants.PATH_ADDON_OPTIONS

    data_env = {}

    if not path.exists():
        return data_env

    data_json = path.read_json()

    with logger.span(f'Converting addon "{path}" to environment variables...'):
        for key, value in data_json.items():
            key_env = key.upper()
            val_env = str(value)
            logger.debug(f'Converting {key_env}={ELLIPSIS}" to environment variable...')
            data_env[key_env] = val_env

    return data_env
